@@ -4,6 +4,7 @@ Streaming CLI interface for OpenAI's Chat API.
 import json
 import logging
 import os
+import subprocess
 import threading
 import time
 from enum import Enum
@@ -13,6 +14,7 @@ from typing import *
 import fire
 import openai
 import pkg_resources
+import tenacity
 import tiktoken
 from prompt_toolkit import PromptSession
 from prompt_toolkit.auto_suggest import AutoSuggest
@@ -115,7 +117,11 @@ def chat_completion(*args, **kwargs):
         except ValueError as e:
             logger.warning(f"Unable to determine minimum/maximum cost: {e}")
 
-    completion = openai.ChatCompletion.create(*args, **kwargs)
+    completion = tenacity.retry(
+        stop=tenacity.stop_after_attempt(3),
+        wait=tenacity.wait_fixed(1),
+        reraise=True,
+    )(openai.ChatCompletion.create)(*args, **kwargs)
 
     def log_stream_completion_cost(completion):
         completion_tokens = 0
@@ -196,6 +202,10 @@ class ChatGenerator:
         # Add the user's message to the list of messages
         self.add_message("user", user_message)
 
+        return self.get_response(write=write)
+
+
+    def get_response(self, write: Callable[[str], T] = partial(print, end="")) -> Dict[str, U]:
         # Send the messages to the API, accumulate the responses, and print them as they come in
         response_accumulated = None
 
@@ -234,8 +244,48 @@ class ChatGenerator:
             raise
 
         finalize()
+
         return response_accumulated
 
+    def send_exec(
+            self, user_message: str, write: Callable[[str], T] = partial(print, end=""),
+    ) -> List[Dict[str, U]]:
+        # Add the user's message to the list of messages
+        self.add_message("user", user_message)
+
+        responses = []
+        while True:
+            # Get the response from the API
+            response = self.get_response(write=write)
+
+            # Get any commands to run, run them, and append the output to
+            commands = []
+            for line in response["message"]["content"].splitlines():
+                if line.startswith("!"):
+                    commands.append(line[1:])
+
+            if not commands:
+                break
+
+            outputs = []
+            for command in commands:
+                output = subprocess.check_output(command, shell=True).decode("utf-8")
+                outputs.append(output)
+                write(output)
+
+            # Add the output as a system message
+            self.add_message("system", "".join(outputs))
+
+            responses.append(response)
+
+        return responses
+
+
+
+
+
+
+
     def pop(self) -> MutableMapping:
         """
         Pop the last user message.
@@ -495,16 +545,32 @@ def chatcli(
     """
     # If verbose is enabled, print logs
     if verbose:
+        print("Verbose mode is enabled. This may cause formatting issues due to the way chatcli erases and overwrites "
+              "text using ANSI escape sequences.")
         logger.setLevel(logging.INFO)
-        logger.warn("Verbose mode may cause formatting issues due to the way chatcli erases and overwrites text using "
-                    "ANSI escape sequences.")
+        logger.warning(
+            "Verbose mode may cause formatting issues due to the way chatcli erases and overwrites text using "
+            "ANSI escape sequences.")
 
     if model is None:
         # Get the model from the environment variable
         model = os.environ.get("OPENAI_CHAT_MODEL", "gpt-3.5-turbo")
         logger.info(f"Using model {model!r} from environment variable 'OPENAI_CHAT_MODEL'")
 
-    start_messages_default = [{"role": "system", "content": "You are a helpful assistant."}]
+    system_message = (
+        "You are a helpful assistant. "
+        "You can run any command in the terminal by prefixing it with a bang, like this:\n"
+        "> !python -c 'print(\"Hello world!\")'\n"
+        "The output will be displayed in a system message like this:\n"
+        "> Hello world!"
+        "Control will be returned to the assistant after this. The assistant can then choose to continue the conversation, "
+        "run another command, or pass control back to the user by writing an empty message \"\". "
+        "The user will be able to view the output of the command in the system message as well, so the assistant does not repeat "
+        "the output."
+    )
+
+
+    start_messages_default = [{"role": "system", "content": system_message}]
 
     # Print a header
     chatcli_version = pkg_resources.get_distribution("chatcli").version
@@ -628,7 +694,7 @@ def chatcli(
 
             # Send the user's message to the generator
             try:
-                chat.send(user_message, write=smooth_writer.write)
+                chat.send_exec(user_message, write=smooth_writer.write)
             except KeyboardInterrupt:
                 # Wait for the buffer to empty
                 smooth_writer.wait_until_buffer_empty()
@@ -662,5 +728,12 @@ def test_chatgen() -> None:
     )
 
 
-if __name__ == "__main__":
+def main() -> None:
+    """
+    The main function.
+    """
     fire.Fire(chatcli)
+
+
+if __name__ == "__main__":
+    main()
